; assembler functions code
;(read-case-sensitive #t)
(load "registers.scm")

; for multiple destination instructions,
; d = 0 means store to W and d = 1 store to f


(define MY_ADD  ;changes Z and C and returns k + W
  (lambda (k)
    (let ((sum (+ W k)))
      (if (> sum 255)
          (begin (C-set!) sum)
          (begin (C-clear!)
                 (if (zero? sum)
                     (begin (Z-set!) sum)
                     (begin (Z-clear!) sum)))))))

(define ADDLW  ; W + k -> W
  (lambda (k)
    (set! W (bitwise-and 255 (MY_ADD k)))))

(define ADDWF  ; W + f -> W|f
  (lambda (f d)
    (if (zero? d)
        (ADDLW (file-ref f))
        (file-set! f (bitwise-and 255 (MY_ADD (file-ref f)))))))


(define MY_AND ; changes Z and returns k & W
  (lambda (k)
    (let ((data (bitwise-and W k)))
      (if (zero? data)
          (begin (Z-set!) data)
          (begin (Z-clear!) data)))))

(define ANDLW  ; W & k -> W
  (lambda (k)
    (set! W (MY_AND k))))

(define ANDWF  ; W & f -> W|f
  (lambda (f d)
    (if (zero? d)
        (ANDLW (file-ref f))
        (file-set! f (MY_AND (file-ref f))))))


(define BCF  ; bit b in register f is cleared
  (lambda (f b)
    (bit-clear! b f)))   ; registers.scm

(define BSF  ; bit b in register f is set
  (lambda (f b)
    (bit-set! b f)))     ; registers.scm


(define BTFSC  ; bit test, skip next operation if clear
  (lambda (f b)
    (if (is-bit-set? b (file-ref f))
        #t
        (set! skip 1))))

(define BTFSS  ; skip if set
  (lambda (f b)
    (when (is-bit-set? b (file-ref f))
      (set! skip 1))))


(define CALL      ; call subroutine at PC k
  (lambda (k)     ; k is 11 bits
    (begin (push (+ (PC-ref) 1))
           (PC-set! k)
           (set! NO-OP 1))))


(define CLRF      ; clears f and sets Z
  (lambda (f)
    (begin (file-set! f 0)
           (Z-set!))))

(define CLRW      ; clears W, sets Z
  (lambda ()
    (begin (set! W 0)
           (Z-set!))))


; WDT NOT YET SUPPORTED
(define CLRWDT  ; resets the watchdog timer, clears the prescaler and TO PD
  (lambda ()
    (begin ;(set! WDT 0)
      (set! OPTION_REG (bitwise-and OPTION_REG 240))
      (file-set! 3 (bitwise-ior (STATUS) 24)))))   ; STATUS<4:3>


(define COMF   ; contents of register f are complemented, -> W|f
  (lambda (f d)
    (let ((MY_COM (lambda (f)
                    (let ((complement (bitwise-not (file-ref f))))
                      (if (zero? complement)
                          (begin (Z-set!) complement)
                          (begin (Z-clear!) complement))))))
      (if (zero? d)
          (set! W (MY_COM f))
          (file-set! f (MY_COM f))))))


(define MY_DECF
  (lambda (f d ZorNOP notZornotNOP)
    (let ((MY_DEC (lambda (f)
                    (let* ((fref (file-ref f))
                           (deced (- (if (zero? fref) 256 fref) 1)))
                      (if (zero? deced)
                          (begin (ZorNOP) deced)
                          (begin (notZornotNOP) deced))))))
      (if (zero? d)
          (set! W (MY_DEC f))
          (file-set! f (MY_DEC f))))))

(define DECF        ; decrement f, store in W or f
  (lambda (f d)
    (MY_DECF f d Z-set! Z-clear!)))

(define DECFSZ      ; decrement f, skip next if = zero, store in W or f
  (lambda (f d)
    (MY_DECF f d (lambda () (set! skip 1)) void)))


(define GOTO                       ; branch PC to k
  (lambda (k)                      ; k is 11 bits
    (if (= k (PC-ref))
        (set! keep #t)             ; pour les boucles infinies !
        (begin (PC-set! k)
               (set! NO-OP 1)))))


(define MY_INCF
  (lambda (f d ZorNOP notZornotNOP)
    (let ((MY_INC (lambda (f)
                    (let ((inced (+ (file-ref f) 1)))
                      (if (zero? inced)
                          (begin (ZorNOP) inced)
                          (begin (notZornotNOP) inced))))))
      (if (zero? d)
          (set! W (bitwise-and 255 (MY_INC f)))
          (file-set! f (bitwise-and 255 (MY_INC f)))))))

(define INCF       ; increment f, store in W or f
  (lambda (f d)
    (MY_INCF f d Z-set! Z-clear!)))

(define INCFSZ     ; increment f, skip if = zero, store in W or f
  (lambda (f d)
    (MY_INCF f d (lambda () (set! skip 1)) void)))


(define MY_IOR
  (lambda (k)
    (let ((iored (bitwise-ior W k)))
      (if (zero? iored)
          (begin (Z-set!) iored)
          (begin (Z-clear!) iored)))))

(define IORLW      ; inclusive or litteral with W
  (lambda (k)
    (set! W (MY_IOR k))))

(define IORWF      ; inclusive or W with f, store in W or f
  (lambda (f d)
    (if (zero? d)
        (set! W (MY_IOR (file-ref k)))
        (file-set! f (MY_IOR (file-ref k))))))


(define MOVF      ; copy f to W or f
  (lambda (f d)
    (let ((val (file-ref f)) (checkZ (lambda (v) (if (zero? v) (Z-set!) #t))))
      (begin (checkZ val)
             (if (zero? d)
                 (set! W val)
                 (file-set! f val))))))

(define MOVLW     ; put litteral in W
  (lambda (k)
    (set! W k)))

(define MOVWF     ; copy W to f
  (lambda (f)
    (file-set! f W)))


(define NOP
  (lambda () (void)))


(define OPTION     ; deprecated
  (lambda ()
    (file-set! 2 W)))


(define RETFIE     ; return from interrupt
  (lambda ()
    (begin (PC-set! (TOS))   ; pop
           (bit-set! 7 "0B") ; GIE here and interrups flags below
           (set! in-int #f)  ; we leave interrupt
           (file-set! "0B" (bitwise-and (string->number "11111000" 2) (INTCON))))))  ; delete int flags

(define RETLW      ; return with litteral in W
  (lambda (k)
    (begin (set! W k)
           (PC-set! (TOS)))))

(define RETURN
  (lambda ()
    (PC-set! (TOS))))


(define RLF       ; rotate left f, dest -> W or f
  (lambda (f d)
    (let ((val (file-ref f)) (checkC (lambda (x) (if (>= x 128) (C-set!)))))
      (begin (checkC val)
             (if (zero? d)
                 (set! W (bitwise-and 255 (* 2 val)))
                 (file-set! f (bitwise-and 255 (* 2 val))))))))

(define RRF       ; rotate right f, dest -> W or f
  (lambda (f d)
    (let ((val (file-ref f)) (checkC (lambda (x) (if (odd? x) (C-set!) #t))))
      (begin (checkC val)
             (if (zero? d)
                 (set! W (quotient 2 val))
                 (file-set! f (quotient 2 val)))))))


(define SLEEP      ; goto sleep mode.
  (lambda ()
    (begin ;(set! WDT 0)
      (set! OPTION_REG (bitwise-and OPTION_REG 240))
      (bit-set! 4 3)   ; 1 -> TO
      (bit-clear! 3 3)  ; 0 -> PD
      (set! sleeping #t)))) ; (PC-ref)))))


(define MY_SUB
  (lambda (a b)
    (cond ((= a b) (begin (C-set!) (Z-set!)))
          ((> a b) (begin (C-clear!) (Z-clear!)))
          (else (begin (C-set!) (Z-clear!))))))


(define SUBLW     ; substract W from litteral
  (lambda (k)
    (begin (MY_SUB W k)
           (set! W (- k W)))))

(define SUBWF    ; substract W from f, store in W or f
  (lambda (f d)
    (let ((fref (file-ref f)))
      (begin (MY_SUB W fref)
             (if (zero? d)
                 (set! W (- fref W))
                 (file-set! f (- fref W)))))))


(define SWAPF       ; swap bits in f, <7:4> to <3:0> and  <3:0> to <7:4>
  (lambda (f d)
    (let ((swap (lambda (x) (let ((low (bitwise-and x 15)))
                              (+ (quotient (bitwise-and x 240) 16)
                                 (* low 16))))))
      (if (zero? d)
          (set! W (swap (file-ref f)))
          (file-set! f (swap (file-ref f)))))))



(define TRIS  ; load TRIS register from W, f is PORTA (05) or PORTB (06)
  (lambda (f)
    (if (= f 5)
        (set! TRISA W)
        (set! TRISB W))))


(define MY_XOR
  (lambda (a b)
    (let ((xored (bitwise-xor a b)))
      (if (= xored 0)
          (begin (Z-set!) xored)
          (begin (Z-clear!) xored)))))

(define XORLW     ; exclusive or litteral and W
  (lambda (k)
    (set! W (MY_XOR W k))))

(define XORWF     ; exclusive or f and W; store in W or f.
  (lambda (f d)
    (if (zero? d)
        (set! W (MY_XOR W (file-ref f)))
        (file-set! f (MY_XOR W (file-ref f))))))
