(load "opcode.scm")

(define prog-example
  #((movlw 0) (tris 6) (clrf 6) (clrf 12) (movf 12 0) (movwf 6) (call 9) (incf 12 1) (goto 4) (movlw 31) (movwf 14) (movlw 31) (movwf 13) (decfsz 13 1) (goto 13) (decfsz 14 1) (goto 11) (return) (end)))

(define hex-example "cc00066061818c20300868024a8ca01331f023b31f0234b8da034b8ea02c008")

(define hex2scheme      ; creates and returns a scheme PIC program (same structure that compiled by the soft)
  (lambda (hexdata)     ; hexdata is the string which represents data from a .hex file.
    (let ((ptr 0) (tree (build-ops-tree)))
      (letrec ((iter (lambda (prog line bindata)
                       (if (string=? "" bindata)
                           (begin (vector-set! prog line (list 'end)) prog)
                           (let* ((bin-op (substring bindata 0 14))
                                  (instr (get-instr-from-bin bin-op tree #f)))
                             (if instr
                                 (begin
                                   (if (= (length instr) 2)
                                       ;here two args, one of 3 for bit pos, and 7 for address
                                       (let ((bit (string->number (substring bin-op 4 7) 2))
                                             (file (string->number (substring bin-op 7) 2)))
                                         (vector-set! prog line (list (car instr) bit file)))
                                       ; here instr from variable arity are described
                                       (cond ((= 0 (cadr instr)) (vector-set! prog line (list (car instr))))
                                             ((= 1 (cadr instr)) ; CLRF and MOVWF (takes a fd arg - 7 bits)
                                              (vector-set! prog line
                                                           (list (car instr) (string->number (substring bin-op 7) 2))))
                                             ((= 2 (cadr instr))
                                              (let ((dest (string->number (substring bin-op 6 7) 2))
                                                    (file (string->number (substring bin-op 7) 2)))
                                                ; one more bit for args : destination of data
                                                (vector-set! prog line (list (car instr) file dest))))
                                             ; 8 and 11 are the size of the uniq argument
                                             ; 8 for a litteral expression
                                             ; 11 for the PC reference (label)
                                             ((= 8 (cadr instr))
                                              (vector-set! prog line
                                                           (list (car instr) (string->number (substring bin-op 7) 2))))
                                             ((= 11 (cadr instr))
                                              (vector-set! prog line
                                                           (list (car instr) (string->number (substring bin-op 4) 2))))
                                             ((= 3 (cadr instr))
                                              ; just for TRIS - arg is 3 bits (or (= arg 5) (= arg 6))
                                              (vector-set! prog line
                                                           (list (car instr) (string->number (substring bin-op 11) 2))))
                                             (else (error "impossible case"))))
                                   (iter prog (+ line 1) (substring bindata 14)))
                                 (error "unknown opcode ! line" line bin-op instr)))))))
        (let* ((bin (number->string (string->number hexdata 16) 2))
               (binplen (string-length bin))
               (plen (/ binplen 14)))
          (if (integer? plen)
              (iter (make-vector (+ plen 1)) 0 bin)
              (error "this program has a bad size, cannot continue")))))))



;;;;;;;;;;;;;;;;; trees functions ;;;;;;;;;;;;;;;;;

(define empty-tree  ; builds an empty tree and returns it
  (lambda ()
    (list #f '() '())))

(define ls         ; left son
  (lambda (tree)
    (cadr tree)))

(define rs         ; right son
  (lambda (tree)
    (caddr tree)))

(define build-tree
  (lambda (value ls rs)
    (list value ls rs)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define build-ops-tree   ; generates a tree from bitOps ByteOps and lit&ctrlOps lists
  (lambda ()             ; needed by hex2scheme
    (letrec ((add (lambda (tree instr binstr)
                    (let ((b (if (string=? binstr "") "" (substring binstr 0 1))))
                      (cond ((string=? "0" b) (build-tree (car tree)
                                                          (add (if (null? (ls tree))
                                                                   (empty-tree)
                                                                   (ls tree))
                                                               instr (substring binstr 1))
                                                          (rs tree)))
                            
                            ((string=? "1" b) (build-tree (car tree)
                                                          (ls tree)
                                                          (add (if (null? (rs tree))
                                                                   (empty-tree)
                                                                   (rs tree))
                                                               instr (substring binstr 1))))
                            
                            (else (build-tree instr (ls tree) (rs tree))))))))
      
      (let ((global-tree (empty-tree)))
        (for-each (lambda (instr) (set! global-tree (add global-tree instr (caddr instr))))
                  ByteOps)
        (for-each (lambda (instr) (set! global-tree (add global-tree instr (caddr instr))))
                  lit&ctrlOps)
        (for-each (lambda (instr) (set! global-tree (add global-tree instr (cadr instr))))
                  bitOps)
        global-tree))))


(define get-instr-from-bin    ; searches the corresponding item in the tree
  (lambda (binstr tree last)
    (cond ((null? tree) last)
          ((and (null? (ls tree)) (null? (rs tree))) (car tree))
          ((string=? "0" (substring binstr 0 1))
           (if (null? (ls tree))
               (if (car tree) (car tree) last)
               (get-instr-from-bin (substring binstr 1)
                                   (ls tree)
                                   (if (car tree) (car tree) last))))
          ((string=? "1" (substring binstr 0 1))
           (if (null? (rs tree))
               (if (car tree) (car tree) last)
               (get-instr-from-bin (substring binstr 1)
                                   (rs tree)
                                   (if (car tree) (car tree) last))))
          (else (error "weird binary string...")))))


