(load "registers.scm")
(load "interface-inputs.scm")
;(load "interface-inputs.scm")

(require (lib "process.ss"))  ; pour le system

(define largeur 1024)
(define largeur-editeur 450)
(define largeur-dessin (- largeur largeur-editeur))
(define hauteur 768)

(define conf:use-parport #f)
(define mon-port-parallele 888)
(define parport mon-port-parallele)

(define frame
  (instantiate frame% ()
    (label "PIC16F84 Simulator")
    (min-width largeur)
    (min-height hauteur)))

(define hpanel-buttons (instantiate horizontal-panel% ()
                         (parent frame) (stretchable-height #f)))

(define hpanel-buttons-draw (instantiate horizontal-panel% ()
                              (parent hpanel-buttons) (stretchable-height #f)))

(instantiate button% () (label "Exit") (parent hpanel-buttons-draw)
  (callback (lambda (b c) (send frame show #f))))

(instantiate button% () (label "INPUTS") (parent hpanel-buttons-draw)
  (callback (lambda (b c)
              (when (not (send inputs-frame is-shown?))
                (send inputs-frame show #t))
              (refresh-i/o))))

(instantiate button% () (label " RUN/STOP ") (parent hpanel-buttons-draw)
  (callback (lambda (b c) (if (and (thread? simul-thread) (thread-running? simul-thread))
                              (kill-thread simul-thread)
                              (if program
                                  (begin (set! sleeping #f) (set! step #f) (set! simul-thread (thread runtime)))
                                  (printf "you should make before run.\n")))))) ; stop
  
(instantiate button% () (label "STEP") (parent hpanel-buttons-draw)
  (callback (lambda (b c)
              (if program
                  (begin (set! step #t)
                         (if (and (thread? simul-thread) (thread-running? simul-thread))
                             (void)
                             (begin (set! sleeping #f) (runtime))))
                  (printf "you should make before run.\n")))))

(instantiate button% () (label "RESET") (parent hpanel-buttons-draw)
  (callback (lambda (b c)
              (set! sleeping #f)
              (file-set! 3 (bitwise-and 24 (file-ref 3)))      ; STATUS
              (file-set! 11 (bitwise-and 254 (file-ref 11)))   ; INTCON
              (file-set! 129 255)   ; OPTION_REG
              (file-set! 133 31)    ; TRISA
              (file-set! 134 255)   ; TRISB
              (PC-set! 0))))


(instantiate slider% () (label "  speed ") (min-value 1) (max-value 100) (style '(horizontal plain))
  (parent hpanel-buttons-draw) (callback (lambda (b c) (set! wait-time (let ((x (send b get-value))) (- 2.08 (* 0.45 (log x)))))))
  (init-value 50) (min-width 180) (stretchable-width #f))


(define my-white (make-object color% 255 255 255))
(define my-grey (make-object color% 200 200 200))
(define no-brush (instantiate brush% ("BLACK" 'transparent)))
(define red-brush (instantiate brush% ("RED" 'solid)))
(define grey-brush (instantiate brush% (my-grey 'solid)))

(define font-40 (make-object font% 40 'modern 'normal 'bold #f))
(define font-32 (make-object font% 32 'modern 'normal 'bold #f))
(define font-18 (make-object font% 18 'modern 'normal 'bold #f))
(define font-16 (make-object font% 16 'modern 'normal 'bold #f))
(define font-12 (make-object font% 12 'modern 'normal 'normal #f))
(define font-12-b (make-object font% 12 'modern 'normal 'bold #f))
(define font-8 (make-object font% 8 'modern 'normal 'normal #f))

; fonction qui dessine les registres
; en hexa: SRAM W PC TMR0/WDT   la pile !
; en binaire: STATUS INTCON OPTION_REG PORTB TRISB
; en boutons radio: PORTA TRISA POR(reset)
; en image: sleep, interrupt

(define (draw-registers-boxes dc)
  (define (iter-freg i fin y x)
    (if (= fin i)
        #t
        (begin 
          (send dc draw-text (hexa-string i) x (+ y 8) #f 0)
          (send dc draw-rectangle (+ 20 x) y 50 27)
          (iter-freg (+ i 1) fin (+ y 30) x))))
  
  (define pen (make-object pen%))
  (send pen set-width 3)
  (send dc set-pen pen)
  (send dc set-brush no-brush)
  
  ; file registers
  (send dc set-font font-18)
  (send dc draw-text "FILE REGISTERS" 55 5 #f 0)
  
  (send dc set-font font-12)
  (iter-freg 12 35 40 16)
  (iter-freg 35 58 40 100)
  (iter-freg 58 80 40 184)
  
  ; separator
  (send pen set-width 2)
  (send dc set-pen pen)
  (send dc draw-line 280 0 280 hauteur)
  
  ; nouveaux separateurs
  (send dc draw-line 415 0 415 140)   ; vertical, a cot du W
  (send dc draw-line 415 70 largeur-dessin 70) ; entre PC et TMR0
  (send dc draw-line 280 140 largeur-dessin 140) ; horiz sous TMR0
  (send dc draw-line 280 240 largeur-dessin 240) ; entre STATUS et INTCON
  (send dc draw-line 280 340 largeur-dessin 340) ; entre INTCON et OPTION_REG
  (send dc draw-line 280 440 largeur-dessin 440) ; entre OPTION_REG et TRISB
  (send dc draw-line 280 540 largeur-dessin 540) ; entre TRISB et TRISA
  (send dc draw-line 385 540 385 630) ; entre INT et TRISA
  (send dc draw-line 280 630 largeur-dessin 630) ; en dessous de TRISA
  
  ; W
  (send pen set-width 3)
  (send dc set-pen pen)
  (send dc set-font font-32)
  (send dc draw-text "W" 335 5 #f 0)
  
  ; PC, TMR0
  (send dc set-font font-18)
  (send dc draw-text "PC" 435 28 #f 0)
  (send dc draw-text "TMR0" 425 97 #f 0)
  
  ; binary registers status intcon option_reg trisb trisa
  (send dc set-font font-16)
  (send dc draw-text "STATUS" 380 150 #f 0)
  (send dc draw-text "INTCON" 380 250 #f 0)
  (send dc draw-text "OPTION_REG" 350 350 #f 0)
  (send dc draw-text "TRISB" 390 450 #f 0)
  (send dc draw-text "TRISA" 440 550 #f 0)
  (send dc set-font font-12-b)
  (send dc draw-text "SLEEP/INT" 295 550 #f 0)
  (send dc set-font font-12)
  (send dc draw-text "IRP RP1:0  TO PD  Z  DC C" 330 172 #f 0)
  (send dc set-font font-8)
  (send dc draw-text "GIE EEIE T0IE INTE RBIE T0IF INTF RBIF" 320 272 #f 0)
  (send dc draw-text "RBPU INTEDG T0CS T0SE PSA PS2:0" 324 372 #f 0))


; dessine les leds et leur tat pour ports A(5) et B(8)

(define (iter-ports dc i x y n PORT)
  (if (= i n)
      #t
      (begin (if (is-bit-set? (- n i 1) PORT)
                 (send dc set-brush red-brush)
                 (send dc set-brush grey-brush))
             (send dc draw-ellipse x y 25 25)
             (iter-ports dc (+ i 1) (+ x 30) y n PORT))))


;  affiche PORTB sur le port parallele
(define (output)
  (when conf:use-parport
    (system (string-append "./output " (number->string (PORTB)) " " (number->string parport)))))


; redessine les 2 ports dans la fenetre pour eviter de tout redessiner si juste ca bouge

(define (refresh-ports)
  (iter-ports $bitmap-dc 0 300 650 8 (PORTB))
  (iter-ports $bitmap-dc 0 390 690 5 (PORTA))
  (output))

(define (INT-LED state dc)
  (if (equal? 'off state)
      (begin (send dc set-brush grey-brush)
             (send dc draw-ellipse 340 585 25 25))
      (begin (send dc set-brush red-brush)
             (send dc draw-ellipse 340 585 25 25))))

; fonction qui dessine le contenu des registres et les trucs spciaux
; elle est appele au dbut et a chaque boucle du programme pour mettre a jour.

(define (draw-registers-content dc)
  (define (iter-freg i fin y x)
    (if (= fin i)
        #t
        (begin 
          (send dc draw-text (hexa-string (vector-ref mem (+ i 12))) x (+ y 8) #f 0)
          (iter-freg (+ i 1) fin (+ y 30) x))))
  
  ; file registers
  (send dc set-font font-18)
  (iter-freg 0 23 38 49)
  (iter-freg 23 46 38 133)
  (iter-freg 46 68 38 216)
  
  ; W
  (send dc set-font font-40)
  (send dc draw-text (hexa-string W) 317 57 #f 0)
  
  ; PC & TMR0
  (send dc set-font font-32)
  (send dc draw-text (hexa-string (PC-ref)) 475 13 #f 0)
  (send dc draw-text (hexa-string (TMR0)) 495 83 #f 0)
  
  ; binary registers status intcon option_reg trisb trisa
  (send dc draw-text (binary-string (STATUS) 8) 331 186 #f 0)
  (send dc draw-text (binary-string (INTCON) 8) 331 282 #f 0)
  (send dc draw-text (binary-string OPTION_REG 8) 331 385 #f 0)
  (send dc draw-text (binary-string TRISB 8) 331 475 #f 0)
  (send dc draw-text (binary-string TRISA 5) 400 570 #f 0)
  
  ; PORTA PORTB  -  LEDS 
  (iter-ports dc 0 300 650 8 (PORTB))
  (iter-ports dc 0 390 690 5 (PORTA))
  
  ; SLEEP LED
  (if sleeping
      (send dc set-brush red-brush)
      (send dc set-brush grey-brush))
  (send dc draw-ellipse 300 585 25 25)
  
  ; INT LED
  (INT-LED 'off dc)
  )

(define hpanel-main (instantiate horizontal-panel% () (parent frame)))

(define $bitmap (make-object bitmap% largeur-dessin hauteur))
(define $bitmap-dc (make-object bitmap-dc%))
(send $bitmap-dc set-bitmap $bitmap)
;(send $bitmap-dc clear)

;(send $bitmap-dc set-background my-grey)
(send $bitmap-dc set-text-mode 'solid)
(send $bitmap-dc set-text-background my-white)

(send $bitmap-dc clear)

;-------------------------- DESSIN ICI --------------------------
(draw-registers-boxes $bitmap-dc)
(draw-registers-content $bitmap-dc)



; pour le refresh  -  memento du schemer :)

(define $bitmap-canvas%
  (class* canvas% ()
    (inherit get-dc)
    (define/override on-paint
      (lambda () (send (get-dc) draw-bitmap $bitmap 0 0)))
    (super-instantiate ())))

(define $canvas
  (make-object $bitmap-canvas% hpanel-main))

;(send frame maximize #t)     ; not for X

(define refresh
  (lambda ()
    (draw-registers-content $bitmap-dc)
    (send $canvas refresh)))


(output)