(load "asm_func.scm")
(load "interface-editeur.scm")

; the program is a 1k max words vector. (1024 * 14 bits)
; the line in process is the program counter (PC)
; PC is 11 bit wide, so can call 2k memory, more than the C has.
; lines in the vector are quoted scheme expressions; ex:
; (MOVLW 12) if in dec or (MOVLW "0c") if in hex (2 are supported)
; real PC is (PC-ref), modified by CALL RETURN GOTO RETLW RETFIE
; labels for GOTO and CALL are given in a seperate AList ((name . addr) ...)
; addr is here the PC of the label.

; 0 -> W
; 1 -> f

;(define program             ;  programme qui compte - pour le debug
;  (vector '(MOVLW 0)
;          '(TRIS 6)
;          '(CLRF 6)
;          '(CLRF "0c")
;          '(MOVF "0C" 0)
;          '(MOVWF 6)
;          '(INCFSZ "0C" 1)
;          '(GOTO 4)
;          '(RETURN)))

(define program #f)

; variables needed for simulation
; these two next depends from the prescaler
;(define TMR0-cycles 0)   ; count number of cycles since last inc of TMR0
;(define WDT-cycles 0)    ; same for the WatchDog Timer

; WDT base timeout is 18ms so period is 70.3125 s

(define simul-thread #f)
(define step #f)
(define keep #f)            ; pour les boucles infinies.
(define wait-time 0.32)     ; seconds   (50% of the slider)
(define old-PORTB (PORTB))
(define in-int #f)

; funtions used in runtime

(define run eval)

; interrupts are:
; if Global Interrupt Bit is set (GIE INTCON<7>)
; TMR0 overflow  INTCON<5> (enabled)  INTCON<2> (has occured)
; INTCON<5> selects wether TMR0 is clock (internal) or counter (RA4)
; RB0/INT INTCON<4> (enabled)  INTCON<1> (has occured)  ->
;   -> OPTION_REG<6> (edge)   ? OPTION_REG<7> (enabled)
; RB7:RB4 (when input) INTCON<3> (enabled) INTCON<0> (has occured)
; data eeprom write complete - not implemented
;
; when an interrup occurs:  NOT DONE !
; clear GIE bit
; push PC in stack
; new PC addr is 004
; GIE is set and flags are reset in RETFIE

(define catch-interrupt     ; returns boolean
  (lambda ()
    (file-set! 11 (bitwise-and 248 (file-ref 11)))  ; pour mettre les flags a 0.
    (let ((catched
           (cond ((= 255 (TMR0))
                  (begin (bit-set! 2 11)      ;  T0IF
                         (if (is-bit-set? 5 (INTCON)) 'TMR0 #f)))
                 ((not (= (bitwise-and (PORTB) 241) (bitwise-and old-PORTB 241)))   ; ints on portb
                  (if (or (and (is-bit-set? 0 (PORTB)) (not (is-bit-set? 0 old-PORTB))
                               (is-bit-set? 6 OPTION_REG))
                          (and (is-bit-set? 0 old-PORTB) (not (is-bit-set? 0 (PORTB)))
                               (not (is-bit-set? 6 OPTION_REG))))
                      (begin (bit-set! 1 11) (if (is-bit-set? 4 (INTCON)) 'RB0/INT #f))
                      ;else -> bits PORTB<7:4> changed
                      (let ((portbchange
                             (cond ((= 0 (bitwise-and 240 TRISB)) #f)   ; not needed, just for speed
                                   ((and (is-bit-set? 4 TRISB)
                                         (not (= (bitwise-and (PORTB) 16) (bitwise-and old-PORTB 16))))
                                    (begin (bit-set! 0 11) 'RB4))
                                   ((and (is-bit-set? 5 TRISB)
                                         (not (= (bitwise-and (PORTB) 32) (bitwise-and old-PORTB 32))))
                                    (begin (bit-set! 0 11) 'RB5))
                                   ((and (is-bit-set? 6 TRISB)
                                         (not (= (bitwise-and (PORTB) 64) (bitwise-and old-PORTB 64))))
                                    (begin (bit-set! 0 11) 'RB6))
                                   ((and (is-bit-set? 7 TRISB)
                                         (not (= (bitwise-and (PORTB) 128) (bitwise-and old-PORTB 128))))
                                    (begin (bit-set! 0 11) 'RB7))
                                   (else #f)))) ; changed is output
                        (if (is-bit-set? 3 (INTCON)) portbchange #f))))
                 ; eeie not implemented.
                 (else #f))))
      ;(printf "catched dans catch-interrupt: ~a\n" catched)
      (set! old-PORTB (PORTB))
      (if catched
          (begin (INT-LED 'on $bitmap-dc)
                 (if (is-bit-set? 7 (INTCON))    ; GIE is set
                     catched
                     #f))
          (begin (INT-LED 'off $bitmap-dc)
                 #f)))))


; this part is for a later version...
; Prescaler Rate - bits OPTION_REG<0:2>
; value | TMR0 rate |  WDT rate
;--------------------------------
;  000  |    1:2    |    1:1
;  001  |    1:4    |    1:2
;  010  |    1:8    |    1:4
;  011  |    1:16   |    1:8
;  100  |    1:32   |    1:16
;  101  |    1:64   |    1:32
;  110  |    1:128  |    1:64
;  111  |    1:256  |    1:128
;
; WDT is always counting, even in SLEEP mode
; TMR0 has a read/write access by user. if it is write,
; it takes 2 cycles to resync with internal clocks.
; in-TMR0-or-WDT is executed each cycle
; the prescaler is not used for WDT and TMR0 in the same time,
; OPTION_REG<3> tells which is used.

;(define inc-TMR0-and-WDT
;  (lambda ()
;    (let ((select (bitwise-and OPTION_REG 8)) (value (bitwise-and OPTION_REG 7)))
;      (if (= select 1)
;          (begin (if (<= (- (expt 2 value) 1) WDT-cycles)
;                     (begin (set! WDT (+ WDT 1))
;                            (set! WDT-cycles 0))
;                     (set! WDT-cycles (+ WDT-cycles 1)))
;                 (set! TMR0 (+ TMR0 1)))
;          (begin (if (<= (- (expt 2 (+ value 1)) 1) TMR0-cycles)
;                     (begin (set! TMR0 (+ TMR0 1))
;                            (set! TMR0-cycles 0))
;                     (set! TMR0-cycles (+ TMR0-cycles 1)))
;                 (set! WDT (+ WDT 1)))))))         ; NOT SURE !


(define end    ; keyword at the end of programs
  (lambda ()
    (if (thread-running? simul-thread)
        (kill-thread simul-thread)
        (PC-set! 0))))


; must check or process this during runtime:
; skleeping NO-OP TMR0 PC skip
; PC is modified by asm fuction.
; (1) after the wake-up, instruction after the SLEEP is executed, and then
; PC+1 is pushed and PC is set to 004
; This is the main loop of the simulation
; INT FLAGS MUST BE SET NOT DEPENDING THEIR ENABLING BIT
; TODO:
; le sleep
; doc: finir le code par un \n
;
; (send a-text:basic highlight-range start end color bitmap)
; (send a-text last-line) -> exact non-negative integer
; (send a-text line-start-position line visible?) -> exact non-negative integer

(define runtime
  (lambda ()
    (let ((PC (PC-ref)))
      (refresh)
      (if (not (= old-PORTB (PORTB)))
          (output))
      (sleep/yield wait-time)
      (when (not sleeping)
        (file-set! 1 (bitwise-and 255 (+ (TMR0) 1))))   ; (modulo 256)
      (let ((catched (if in-int #f (catch-interrupt))))
        ;(bit-clear! 4 3)     ; TO   pour le sleep wakeup par wdt
        (cond (catched (begin (printf "!interrupt!: ~a\n" catched)
                              (bit-clear! 7 3)     ; IRP - use 00-ff databank
                              (bit-clear! 7 11)    ; disable GIE
                              (push (PC-ref))      ; save where we were
                              (PC-set! 004)        ; branch to 004
                              (set! in-int #t)     ; we go in interrupt
                              (set! NO-OP 3)       ; idle 3 loops
                              (set! keep #t)))     ; and the fourth final idle
              ((>= NO-OP 1) (set! NO-OP (- NO-OP 1)))
              ((= skip 1) (begin (PC-set! (+ (PC-ref) 1)) (set! skip 0)))
              (sleeping #t)
              (else (begin (run (vector-ref program (PC-ref)))
                           (if (= PC (PC-ref))
                               (if keep
                                   (set! keep #f)
                                   (PC-set! (+ (PC-ref) 1))))))))
      (if step
          (void)
          (runtime)))))


(define (lower-case-char char)
  (integer->char (let ((i (char->integer char)))
                   (if (<= 65 i 90) (+ i 32) i))))

(define (lc str)     ; lower-case-string
  (list->string (map lower-case-char (string->list str))))


; construction de la A-liste des labels, et de celle des constantes
; crer un tableau de la taille du nombre de lignes du fichier,
; qui fait correspondance entre la ligne du programme et la ligne dans le fichier.

(define pre-compile
  (lambda ()
    (let* ((line-count (send my-text last-line))
           (PC<->file-line (make-vector line-count -1)))
      (letrec ((iter
                (lambda (file-line prog-line labels symbols)
                  (if (> file-line line-count)
                      (begin (set! program (make-vector prog-line 0))
                             (printf "\nlabels ~a\nsymbols ~a\nindices ~a\n" labels symbols PC<->file-line)
                             (values prog-line labels symbols PC<->file-line))       ; FIN
                      (let* ((start (send my-text line-start-position file-line))
                             (strlen (send my-text line-length file-line))
                             (line-string (send my-text get-text start (+ start strlen -1) #f #f)))
                        
                        (if (string=? "" line-string)
                            (iter (+ file-line 1) prog-line labels symbols)          ; ligne vide
                            
                            (let ((exploded (regexp-match* "[^ |\t]+" line-string))) ; on coupe en mots
                              (if (char=? (string-ref (car exploded) 0) #\;)
                                  (iter (+ file-line 1) prog-line labels symbols)    ; ligne commente
                                  
                                  (if (or (char=? #\space (string-ref line-string 0))
                                          (char=? #\tab (string-ref line-string 0)))
                                      ;not a label nor an EQU (la ligne ne commence pas au 1er char)
                                      (if (string-ci=? "org" (car exploded))
                                          (let ((origin (remove-0x (cadr exploded))))
                                            (vector-set! PC<->file-line origin file-line)
                                            (iter (+ file-line 1) origin labels symbols))
                                          ; so it is a real line
                                          (begin (vector-set! PC<->file-line prog-line file-line)
                                                 (iter (+ file-line 1) (+ prog-line 1) labels symbols)))
                                      
                                      ;(if (regexp-match "#define[\t| ]" (car exploded))
                                      ; labels, il faut voir si il est seul sur la ligne ou pas
                                      (if (or (= 1 (length exploded))
                                              (char=? (string-ref (cadr exploded) 0) #\;))
                                          (iter (+ file-line 1) prog-line           ; a label alone
                                                (cons (cons (car exploded) prog-line) labels)
                                                symbols)
                                          (if (string-ci=? "equ" (cadr exploded))   ; EQU
                                              (iter (+ file-line 1) prog-line labels
                                                    (cons (cons (car exploded) (remove-0x (caddr exploded))) symbols))
                                              ; a label and his code
                                              (begin (vector-set! PC<->file-line prog-line file-line)
                                                     (iter (+ file-line 1) (+ prog-line 1)
                                                           (cons (cons (car exploded) prog-line) labels)
                                                           symbols)))))))))))))
        (iter 0 0 '() '())))))


; (send a-text get-text start end flattened? force-cr? )-> string 
; (send a-text line-start-position line visible? )-> exact non-negative integer
; (send a-text find-wordbreak start end reason )-> void 
; (send a-text move-position code extend? kind )-> void


; transformation du fichier en expressions scheme.
; on a les symboles les labels et les vraies lignes grace a pre-compile.

(define compile
  (lambda ()
    (let-values (((prog-line-count labels symbols PC<->file-line) (pre-compile)))
      (letrec ((iter
                (lambda (prog-line)
                  (if (= prog-line prog-line-count)
                      (begin (replace-0-NOP prog-line) (display program) (newline))
                      (if (= -1 (vector-ref PC<->file-line prog-line))
                          (iter (+ 1 prog-line))
                          ;(set! prog-line (+ 1 prog-line))
                          (let* ((start (send my-text line-start-position (vector-ref PC<->file-line prog-line)))
                                 (strlen (send my-text line-length (vector-ref PC<->file-line prog-line)))
                                 (file-line (send my-text get-text start (+ start strlen -1) #f #f))  ; ligne du fichier
                                 (no-comment (regexp-replace "([^;]+)([ ]+;.*)" file-line "\\1"))     ; pareil sans les commentaires
                                 (exploded (regexp-match* "[^, |\t]+" no-comment))                    ; pareil2 coup en mots
                                 (nbmots (length exploded)))                                          ; nombre de mots
                            (when (assoc (car exploded) labels)     ; cette ligne commence par un label
                              (set! exploded (cdr exploded))        ; on commence alors au deuxieme mot
                              (set! nbmots (- nbmots 1)))           ; et on dit qu'il y en a un de moins donc
                            
                            (cond ((or (string-ci=? "call" (car exploded))      ; les seules qui utilisent des labels.
                                       (string-ci=? "goto" (car exploded)))
                                   (vector-set! program prog-line
                                                (list (string->symbol (lc (car exploded))) (cdr (assoc (cadr exploded) labels)))))
                                  
                                  ((= 1 nbmots) (vector-set! program prog-line (list (string->symbol (lc (car exploded))))))
                                  (else (let ((symbol (assoc (cadr exploded) symbols)))
                                          (cond ((and (= 2 nbmots) symbol)
                                                 (vector-set! program prog-line (list (string->symbol (lc (car exploded))) (cdr symbol))))
                                                ((and (= 2 nbmots) (char=? #\b (string-ref (cadr exploded) 0))
                                                      (char=? #\'(string-ref (cadr exploded) 1)))     ; b'10101010'
                                                 (vector-set! program prog-line (list (string->symbol (lc (car exploded)))
                                                                                      (string->number (substring (cadr exploded) 2 10) 2))))
                                                ((= 2 nbmots)    ; l'argument est un nombre normal
                                                 (vector-set! program prog-line (list (string->symbol (lc (car exploded)))
                                                                                      (remove-0x (cadr exploded)))))
                                                ((= 3 nbmots) (let ((dest (cond ((or (string-ci=? "w" (caddr exploded))
                                                                                     (string=? "0"(caddr exploded))) 0)
                                                                                ((or (string-ci=? "f" (caddr exploded))
                                                                                     (string-ci=? "1" (caddr exploded))) 1)
                                                                                (else (string->number (caddr exploded)))))) ; n de bit
                                                                (if symbol
                                                                    (vector-set! program prog-line (list (string->symbol (lc (car exploded)))
                                                                                                         (cdr symbol) dest))
                                                                    (vector-set! program prog-line (list (string->symbol (lc (car exploded)))
                                                                                                         (remove-0x (cadr exploded)) dest)))))
                                                (else (printf "erreur: ~a : ~a\n" no-comment nbmots)
                                                      (vector-set! program prog-line (list 'NOP)))))))
                            (iter (+ 1 prog-line))))))))
        (iter 0)))))


(define replace-0-NOP    ; with ORG uses, there may have 0 in the program
  (lambda (n)
    (letrec ((replace (lambda (i)
                        (if (= i n)
                            (void)
                            (begin (when (number? (vector-ref program i))  ; sure it is 0
                                     (vector-set! program i (list 'NOP)))
                                   (replace (+ i 1)))))))
      (replace 0))))
